@@ -1,3 +1,25 @@
+# Copyright 2021 Charilaos K. Mylonas
+# 
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#     http://www.apache.org/licenses/LICENSE-2.0
+# 
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+
+# Notes:
+# -------
+# This code contains the main logic for the full GraphNet computation. 
+# Some ideas (and sometimes source code) were taken from
+# here: https://github.com/deepmind/graph_nets/blob/64771dff0d74ca8e77b1f1dcd5a7d26634356d61/graph_nets/
+# 
+
+
 # python imports
 import os
 from enum import Enum
@@ -17,6 +39,77 @@ import numpy as np
 
 from .datastructures import Graph
 
+
+# The following stores how larger is the aggregation operation 
+# compared to the message output. For instance mean_max is 
+# 2 x [dimension of incoming messages] 
+AGG_TO_MESSAGE_DICT = {
+                        'mean'             : 1, 
+                        'max'              : 1, 
+                        'min'              : 1, 
+                        'sum'              : 1, 
+                        'mean_max'         : 2,
+                        'mean_max_min'     : 3,
+                        'mean_max_min_sum' : 4}
+
+
+
+
+def _unsorted_segment_reduction_or_zero(reducer, values, indices, num_groups):
+  """
+  Common code for unsorted_segment_{min,max}_or_zero (below).
+  """
+
+  reduced = reducer(values, indices, num_groups)
+  present_indices = tf.math.unsorted_segment_max(
+      tf.ones_like(indices, dtype=reduced.dtype), indices, num_groups)
+  present_indices = tf.clip_by_value(present_indices, 0, 1)
+  present_indices = tf.reshape(
+      present_indices, [num_groups] + [1] * (reduced.shape.ndims - 1))
+  reduced *= present_indices
+  return reduced
+
+
+def unsorted_segment_min_or_zero(values, indices, num_groups,
+                                 name="unsorted_segment_min_or_zero"):
+  """Aggregates information using elementwise min.
+  Segments with no elements are given a "min" of zero instead of the most
+  positive finite value possible (which is what `tf.math.unsorted_segment_min`
+  would do).
+  Args:
+    values: A `Tensor` of per-element features.
+    indices: A 1-D `Tensor` whose length is equal to `values`' first dimension.
+    num_groups: A `Tensor`.
+    name: (string, optional) A name for the operation.
+  Returns:
+    A `Tensor` of the same type as `values`.
+  """
+  with tf.name_scope(name):
+    return _unsorted_segment_reduction_or_zero(
+        tf.math.unsorted_segment_min, values, indices, num_groups)
+
+
+def unsorted_segment_max_or_zero(values, indices, num_groups,
+                                 name="unsorted_segment_max_or_zero"):
+  """Aggregates information using elementwise max.
+  Segments with no elements are given a "max" of zero instead of the most
+  negative finite value possible (which is what `tf.math.unsorted_segment_max`
+  would do).
+  Args:
+    values: A `Tensor` of per-element features.
+    indices: A 1-D `Tensor` whose length is equal to `values`' first dimension.
+    num_groups: A `Tensor`.
+    name: (string, optional) A name for the operation.
+  Returns:
+    A `Tensor` of the same type as `values`.
+  """
+  with tf.name_scope(name):
+    return _unsorted_segment_reduction_or_zero(
+        tf.math.unsorted_segment_max, values, indices, num_groups)
+
+
+
+
 # Loading some self-assets:
 try:
     html_asset_path = os.environ['TFGNNS_HTML_ASSETS']
@@ -24,8 +117,8 @@ try:
         ACCORDION_CSS = f.read()
     with open(os.path.join(html_asset_path,'accordion.js')) as f:
         ACCORDION_JS  = f.read()
-except e:
-    print("Some assets failed to load!")
+except:
+    print("Some assets failed to load!\nSome pretty printing functionality may be unavailable.")
     
 
 def _instantiate_gamma(t, NParams_ = 1):
@@ -57,8 +150,6 @@ def _split_tensor_dict_to_repar(td, nlatent_nodes_or_all, nlatent_edges = None,
     return td_mean, td_std
 
 
-
-
 EDGE_FUNCTION_INPUTS = ['global_state','sender_node_state','receiver_node_state','edge_state']
 class EdgeInput(Enum):
     GLOBAL_STATE = 'global_state'
@@ -116,7 +207,7 @@ class GraphNet:
                                         used for the edges-to-nodes aggregation and edges-to-global. 
 
             node_to_global_aggregation_function : the node aggregator. This aggregates all nodes to provide
-                                        an input to the global MLP.
+                                        an input to the global MLP. Possible values:['mean','max','min', 'sum','mean_max_min']
 
             use_global_input          : whether to use the global input or not
 
@@ -171,8 +262,6 @@ class GraphNet:
             else:
                 self.edge_aggregation_function = edge_aggregation_function        
 
-            #self.scan_edge_to_node_aggregation_function(node_function)
-
         if node_to_global_aggregation_function is not None:
             try:
                 len_ea = len(node_to_global_aggregation_function)
@@ -180,12 +269,6 @@ class GraphNet:
             except:
                 None
 
-
-        
-        #self.has_seg_aggregator_node_to_edge = False
-        #self.has_seg_aggregator_node_to_global = False
-
-        
         if self.edge_function is not None: 
             self.edge_input_size = self.edge_function.inputs[0].shape[1] # input dimension 1 of edge mlp is the edge state size by convention.
 
@@ -231,7 +314,7 @@ class GraphNet:
 
 
 
-        #These are booleans so TF can easilly make tf.functions and not re-trace them.
+        # The following help TF to avoid retracing tf.functions blocks. 
         self.uses_edge_state, self.uses_sender_node_state, self.uses_receiver_node_state = [False, False, False]
 
         if EdgeInput.EDGE_STATE.value in self.edge_input_dict.keys():
@@ -312,6 +395,9 @@ class GraphNet:
             print("* No aggregation function - Graph-independent network.")
 
     def _repr_html_(self):
+        """
+        Print HTML buttons - for jupyter. 
+        """
         s = ''
         s +=  '<div>'
         #s += '<html>'
@@ -399,8 +485,11 @@ class GraphNet:
         # 
         # Global state can also be supplied as an argument (if the edge and/or node functions use it)
         #
-        # parameters:
-        #  tf_graph_tuple : a GraphTuple object containing nodes, edges and their connectivity for multiple graphs.
+        # Arguments:
+        #  tf_graph_tuple : a GraphTuple object (i.e. a dictionary containing tensors) containing nodes, edges and their connectivity for multiple graphs.
+        # 
+        # Outputs:
+        #  A 
 
         # if the graph is not graph_indep compute the edge-messages with the aggregator.
 
@@ -453,8 +542,6 @@ class GraphNet:
 
         # 4) Global computation:
         # Aggregate edges to global:
-
-
         global_inputs = {}
         if (GlobalInput.EDGE_AGG_STATE.value in self.global_input_dict.keys()):
             if self.has_seg_aggregator_node_to_edge: #<- same aggregator for edges-to-nodes and edges-to-global.
@@ -517,8 +604,7 @@ class GraphNet:
             node_inputs.update({NodeInput.NODE_STATE.value : nodes})
 
         if NodeInput.GLOBAL_STATE.value in self.node_input_dict.keys():
-            node_reps = _global_reps_for_nodes
-            node_inputs.update({NodeInput.GLOBAL_STATE.value : tf.gather(global_attr,node_reps)})
+            node_inputs.update({NodeInput.GLOBAL_STATE.value : tf.gather(global_attr,_global_reps_for_nodes)})
 
         if NodeInput.EDGE_AGG_STATE.value in self.node_input_dict.keys():
             if self.has_seg_aggregator_edge_to_global:
@@ -573,7 +659,7 @@ class GraphNet:
           'edges','nodes','senders','receivers','n_edges','n_nodes','global_attr','_global_reps_for_edges','_global_reps_for_nodes'
 
         """
-        d_ = d.copy() # Not sure if this harms performance. 
+        d_ = d.copy() # A working on a copy of d (for tf.function compilation requirement).
         if self.edge_function is not None:
             new_edges = self.edge_block(**d_)
             d_['edges'] = new_edges
@@ -654,7 +740,7 @@ class GraphNet:
                             'node_state' : n.node_attr_tensor, 
                             'edge_state_agg' : edge_to_node_agg
                         }) # TODO: Change that to infer inputs from function names like the GraphTuple method.
-                                                                                              #       Also treat "globals" with this method (currently un-treated)
+                                                                                              
                 n.set_tensor(node_attr_tensor)
 
         if return_messages:
@@ -663,6 +749,12 @@ class GraphNet:
         return graph
 
     def save(self, path):
+        """
+        Save the model. Iterates of the keras models required and saves them in a folder.
+
+        Arguments:
+          path: the path to save to. Creates it if it does not exist.
+        """
         functions = [self.node_function, self.edge_aggregation_function, self.edge_function]
         path_labels = ["node_function", "edge_aggregation_function", "edge_function"]
         import os
@@ -959,6 +1051,11 @@ def make_mlp(units, input_tensor_list , output_shape, activation = "relu", **kwa
         act_last_layer = kwargs['activate_last_layer']
     else:
         act_last_layer = False
+
+    if 'layernorm_last_layer' in kwargs.keys():
+        layernorm_last_layer = kwargs['layernorm_last_layer']
+    else:
+        layernorm_last_layer = False
     
     
     # A workaround to keep track of created weights easily:
@@ -973,16 +1070,31 @@ def make_mlp(units, input_tensor_list , output_shape, activation = "relu", **kwa
                 return Dense(*args, **kwargs)
 
     dense_maker = DenseMaker()
-    y = dense_maker.make(units, use_bias = False)(edge_function_input)
-    y = dense_maker.make(units, activation = activation)(y)
-    y = dense_maker.make(units, activation = activation)(y)
+    if not isinstance(units,list):
+        # If the input is not a list, it is assumed to be an integer.
+        # Then a two-layer MLP is constructed.
+        y = dense_maker.make(units, use_bias = False)(edge_function_input)
+        y = dense_maker.make(units, activation = activation)(y)
+        y = dense_maker.make(units, activation = activation)(y)
+
+        if act_last_layer:
+            y = dense_maker.make(output_shape[0], activation = activation)(y)
+        else:
+            y = dense_maker.make(output_shape[0])(y)
 
-    if act_last_layer:
-        #y = tf.keras.layers.LayerNormalization()(y)
-        y = dense_maker.make(output_shape[0], activation = activation)(y)
     else:
-        #y = tf.keras.layers.LayerNormalization()(y)
-        y = dense_maker.make(output_shape[0])(y)
+        if len(units) == 0:
+            y = dense_maker.make(output_shape[0], use_bias = False)(edge_function_input)
+        else:
+            y = edge_function_input
+            for kk, u in enumerate(units):
+                if kk == 0:
+                    y = dense_maker.make(u, use_bias = False)(y)
+                else:
+                    y = dense_maker.make(u)(y)
+
+    if layernorm_last_layer:
+        y = tf.keras.layers.LayerNormalization()(y)
         
     if 'model_name' in kwargs.keys():
         name= kwargs['model_name']
@@ -1076,7 +1188,6 @@ def make_global_mlp(units, global_in_size = None,
 
     with tf.name_scope("glob_fn") as scope:
         kwargs['model_name'] = scope + 'model'
-        #print("scope: %s, global_emb_size : %i"%(scope,global_emb_size[0]))
         return make_mlp(units, global_inputs_list , global_emb_size, activation = activation, **kwargs)
 
 
@@ -1146,8 +1257,8 @@ def make_keras_simple_agg(input_size, agg_type):
     dict_agg = {
             'mean' : (tf.reduce_mean,tf.math.unsorted_segment_mean),
             'sum'  : (tf.reduce_sum,tf.math.unsorted_segment_sum),
-            'max'  : (tf.reduce_max,tf.math.unsorted_segment_max),
-            'min'  : (tf.reduce_min,tf.math.unsorted_segment_min)}
+            'max'  : (tf.reduce_max,unsorted_segment_max_or_zero),
+            'min'  : (tf.reduce_min,unsorted_segment_min_or_zero)}
 
     x = Input(shape = input_size, name = "edge_messages") # for "segment" aggregators this needs also a bunch of indices!
     aggs = dict_agg[agg_type]
@@ -1158,7 +1269,7 @@ def make_keras_simple_agg(input_size, agg_type):
     
     return m_basic, m_seg
 
-# TODO: Make the aggregators composable!
+# TODO: Make the aggregators composable
 def make_mean_max_agg(input_size):
     """
     A mean and a max aggregator appended together. This was is useful for some special use-cases.
@@ -1194,7 +1305,7 @@ def make_mean_max_min_agg(input_size):
     m_basic = tf.keras.Model(inputs = x , outputs = agg_out, name = 'basic_meanmaxmin_aggregator')
 
     def tf_function_agg(x, recv, max_seq):
-        v1,v2,v3 = tf.math.unsorted_segment_mean(x,recv,max_seq), tf.math.unsorted_segment_max(x,recv, max_seq) , tf.math.unsorted_segment_min(x, recv,max_seq)
+        v1,v2,v3 = tf.math.unsorted_segment_mean(x,recv,max_seq), unsorted_segment_max_or_zero(x,recv, max_seq) , unsorted_segment_min_or_zero(x, recv,max_seq)
         agg_ss = tf.concat([v1,v2, v3], axis = -1)
         return agg_ss
 
@@ -1244,17 +1355,24 @@ def _aggregation_function_factory(input_shape, agg_type = 'mean'):
     """
     # Each lambda creates a "basic" and a "segment" (sparse) aggregator.
     agg_type_dict = {
-            'mean' : lambda input_shape : make_keras_simple_agg(input_shape, 'mean'),
-            'sum' : lambda input_shape : make_keras_simple_agg(input_shape, 'sum'),
-            'max' : lambda input_shape : make_keras_simple_agg(input_shape, 'max'),
-            'min' : lambda input_shape : make_keras_simple_agg(input_shape, 'min'),
-            'mean_max' : lambda input_shape : make_mean_max_agg(input_shape), # compound aggregator (appending aggregators)
-            'mean_max_min' : lambda input_shape : make_mean_max_min_agg(input_shape),
+            'mean' : lambda input_shape :             make_keras_simple_agg(input_shape, 'mean'),
+            'sum' : lambda input_shape :              make_keras_simple_agg(input_shape, 'sum'),
+            'max' : lambda input_shape :              make_keras_simple_agg(input_shape, 'max'),
+            'min' : lambda input_shape :              make_keras_simple_agg(input_shape, 'min'),
+            'mean_max' : lambda input_shape :         make_mean_max_agg(input_shape), # compound aggregator (appending aggregators)
+            'mean_max_min' : lambda input_shape :     make_mean_max_min_agg(input_shape),
             'mean_max_min_sum' : lambda input_shape : make_mean_max_min_sum_agg(input_shape)
             }
 
-    aggregators = agg_type_dict[agg_type](input_shape)
-    return aggregators
+    
+    try:
+        aggregators = agg_type_dict[agg_type](input_shape)
+        return aggregators
+    except TypeError:
+        aggregators = agg_type_dict[agg_type]((input_shape,))
+        return aggregators
+
+
 
 
 def make_mlp_graphnet_functions(units,
@@ -1467,6 +1585,7 @@ def make_graph_to_graph_and_global_functions(units,
                                             node_or_core_output_size = None, 
                                             edge_output_size = None, 
                                             edge_input_size = None,
+                                            aggregation_function = 'mean',
                                             **kwargs):
     """
     A wrapper to a more general function factory.
@@ -1520,6 +1639,7 @@ def make_graph_to_graph_and_global_functions(units,
                                        use_global_to_edge = False, # <- refers to the global input (that is un-used)
                                        use_global_to_node = False, # <-     >>   >>  >> 
                                        graph_indep=False,
+                                       aggregation_function = aggregation_function,
                                        **kwargs)
 
 
@@ -1530,6 +1650,7 @@ def make_graph_indep_graphnet_functions(units,
         edge_output_size = None,
         global_input_size = None,
         global_output_size = None,
+        aggregation_function = 'mean',
         **kwargs):
 
     """
@@ -1585,6 +1706,7 @@ def make_graph_indep_graphnet_functions(units,
                                         use_global_to_node=False,
                                         create_global_function=True, 
                                         graph_indep=True,
+                                        aggregation_function = aggregation_function,
                                         **kwargs)
 
 
@@ -1596,6 +1718,7 @@ def make_full_graphnet_functions(units,
         edge_output_size = None, 
         global_input_size = None, 
         global_output_size = None, 
+        aggregation_function ='mean',
         **kwargs):
     """
     A wrapper that creates the functions that are needed for a full GN block. 
@@ -1648,7 +1771,9 @@ def make_full_graphnet_functions(units,
                                         use_global_input = True,
                                         use_global_to_edge=True,
                                         use_global_to_node=True,
-                                        create_global_function=True, **kwargs)
+                                        create_global_function=True,
+                                        aggregation_function = aggregation_function,
+                                        **kwargs)
 
 
 
